{***********************************************************************
Unit gfx_colors.PAS v1.5 0801
    (c) by Andreas Moser, amoser@amoser.de

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_colors is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    This unit of gfx_lib is, in part, based on the work of Anders Melander, anders@melander.dk
    (especially the color reduction algorythm)

********************************************************************************}

unit gfx_colors;
{$define DIB_INTFC}

interface
uses
    gfx_basedef,
    gfx_syswrapper,
    gfx_dib,
    gfx_colordither,
    gfx_colorpalette,
    gfx_errors;




 // Image Processing
  procedure GetImagePlane (const ColorPlane:  TColorPlane;
                           const ColorOutput:  BOOLEAN;
                           const OriginalBitmap:  TXBitmap;
                           var   ProcessedBitmap:  TXBitmap);

procedure ReduceColors(SrcBitmap,DstBitmap: TXBitmap; ColorReduction: TColorReduction;
  DitherMode: TDitherMode; ReductionBits: integer; CustomPalette: HxPalette);

//------------------------------------------------------------
// Splitting
//------------------------------------------------------------

 procedure GetRedPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
 procedure GetGreenPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
 procedure GetBluePlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
 procedure GetHuePlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
 procedure GetSatPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
 procedure GetValPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
 procedure GetLigPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);

implementation

 {==  Image Processing  ==============================================}

// -----------------------------------------------------------------------------
//
//			GetImagePlane
//
// -----------------------------------------------------------------------------
procedure GetImagePlane (const ColorPlane:  TColorPlane;
                           const ColorOutput:  BOOLEAN;
                           const OriginalBitmap:  TXBitmap;
                           var   ProcessedBitmap:  TXBitmap);
var
    HSV:THSVTriple;
    CMYK:TCMYKQuad;
    i           :  INTEGER;
    Intensity   :  INTEGER;
    j           :  INTEGER;
  //  L           :  INTEGER;
    RowOriginal :  pRGBarray;
    RowProcessed:  pRGBarray;
begin
    IF   OriginalBitmap.Pixelformat <> xpf24bit
    THEN RAISE  EGraphicformat.Create('GetImageSpace:  ' +
               'Bitmap must be 24-bit color.');

    {Step through each row of image.}
    for j := OriginalBitmap.Height-1 downto 0 DO
    begin
      RowOriginal  := pRGBarray(OriginalBitmap.Scanline[j]);
      RowProcessed := pRGBarray(ProcessedBitmap.Scanline[j]);

      for i := OriginalBitmap.Width-1 downto 0 DO
      begin
        CASE ColorPlane OF
          cpRGB:    IF   ColorOutput
                    THEN AssignRGBTriple(RowProcessed[i], RowOriginal[i])
                    else begin
                      Intensity := RGBIntensity(RowOriginal[i]);
                      AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(Intensity, Intensity, Intensity))
                    end;

          cpRed:    IF   ColorOutput
                    THEN AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(RowOriginal[i].rgbtRed, 0, 0))
                    else begin
                      Intensity := RowOriginal[i].rgbtRed;
                      AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(Intensity, Intensity, Intensity))
                    end;

          cpGreen:  IF   ColorOutput
                    THEN AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(0, RowOriginal[i].rgbtGreen, 0))
                    else begin
                      Intensity := RowOriginal[i].rgbtGreen;
                      AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(Intensity, Intensity, Intensity))
                    end;

          cpBlue:   IF   ColorOutput
                    THEN AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(0, 0, RowOriginal[i].rgbtBlue))
                    else begin
                      Intensity := RowOriginal[i].rgbtBlue;
                      AssignRGBTriple(RowProcessed[i],
                                    RGBTriple(Intensity, Intensity, Intensity))
                    end;

          cpHueHSV:
            begin
              RGBtoHSV(RowOriginal[i], HSV);
              HSV.hsvHue := (255 * HSV.hsvHue) DIV 359;

              {Shades of Hue}
              AssignRGBTriple(RowProcessed[i], RGBTriple(HSV.hsvHue,HSV.hsvHue,HSV.hsvHue))
            end;

          cpSaturationHSV:
            begin
              RGBtoHSV(RowOriginal[i], HSV);

              {Shades of Saturation}
              AssignRGBTriple(RowProcessed[i], RGBTriple(HSV.hsvSaturation,HSV.hsvSaturation,HSV.hsvSaturation))
            end;

          cpValue:
            begin
              RGBtoHSV(RowOriginal[i], HSV);

              {Shades of Value}
              AssignRGBTriple(RowProcessed[i], RGBTriple(HSV.hsvValue,HSV.hsvValue,HSV.hsvValue))
            end;

          cpCyan:
            begin
              RGBtoCMYK(RowOriginal[i], CMYK);

              {Shades of Value}
              AssignRGBTriple(RowProcessed[i], RGBTriple(CMYK.cmykCyan,CMYK.cmykCyan,CMYK.cmykCyan))
            end;

          cpMagenta:
            begin
              RGBtoCMYK(RowOriginal[i], CMYK);

              {Shades of Value}
              AssignRGBTriple(RowProcessed[i], RGBTriple(CMYK.cmykMagenta,CMYK.cmykMagenta,CMYK.cmykMagenta))
            end;

          cpYellow:
            begin
              RGBtoCMYK(RowOriginal[i], CMYK);

              {Shades of Value}
              AssignRGBTriple(RowProcessed[i], RGBTriple(CMYK.cmykYellow,CMYK.cmykYellow,CMYK.cmykYellow))
            end;

          cpBlack:
            begin
              RGBtoCMYK(RowOriginal[i], CMYK);

              {Shades of Value}
              AssignRGBTriple(RowProcessed[i], RGBTriple(CMYK.cmykK,CMYK.cmykK,CMYK.cmykK))
            end;

          cpIntensity:
            begin
              Intensity := RGBIntensity(RowOriginal[i]);

              {Shades of Saturation}
              AssignRGBTriple(RowProcessed[i],
                              RGBTriple(Intensity, Intensity, Intensity))
            end
        end
      end
    end
end;



//------------------------------------------------------------
// Palette
//------------------------------------------------------------


// -----------------------------------------------------------------------------
//
//			Color reduction
//
// -----------------------------------------------------------------------------

procedure ReduceColors(SrcBitmap,DstBitmap: TXBitmap; ColorReduction: TColorReduction;
  DitherMode: TDitherMode; ReductionBits: integer; CustomPalette: HxPalette);
var
  Palette               : HxPalette;
  ColorLookup		: TColorLookup;
  Ditherer		: TDitherEngine;
  Row			: Integer;
  SrcScanLine		,
  Src			: PRGBTriple;
  DstScanLine		,
  Dst			: PChar;
  BGR			: TxRGBTriple;
  BMPIntfc              : TBMPInterface1;
  DIBResult		: TDIBWriter;
  DIBSource		: TDIBReader;

begin
  {$ifdef DIB_INTFC}
  if (ColorReduction = rmNone) then
  begin
      DstBitmap.Assign(SrcBitmap);
      DstBitmap.Pixelformat:=xpf24bit;
      exit;
  end;

  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  if (SrcBitmap.Width*SrcBitmap.Height > BitmapAllocationThreshold) then
  begin
    DstBitmap.Pixelformat:=xpf1bit; // To reduce resource consumption of resize
    BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DstBitmap,xpf1Bit,0,True);
  end
  else

    DIBSource := TDIBReader.Create(SrcBitmap, xpf24bit);
    ColorLookup := nil;
    Ditherer := nil;
    Palette := 0;
    try // Protect above resources

      // Create a palette based on current options
      case (ColorReduction) of
        rmQuantize:
            Palette := doCreateOptimizedPaletteFromSingleBitmap(DIBSource, 1 SHL ReductionBits, 8, False);
        rmQuantizeWindows:
            Palette := CreateOptimizedPaletteFromSingleBitmap(SrcBitmap, 256, 8, True);
        rmPalette:
          Palette := xCopyPalette(CustomPalette);
      else
        Palette := MakePalette(ColorReduction);
      end;

      // Create a color mapper based on current options
          ColorLookup := TColorLookup.Create(Palette,ColorReduction);

      // Nothing to do if palette doesn't contain any colors
      if (ColorLookup.Colors = 0) then
        exit;

      // Create a ditherer based on current options
      case (DitherMode) of
        dmNearest:
          Ditherer := TDitherEngine.Create(SrcBitmap.Width, ColorLookup);
        dmFloydSteinberg:
          Ditherer := TFloydSteinbergDitherer.Create(SrcBitmap.Width, ColorLookup);
        dmStucki:
          Ditherer := TStuckiDitherer.Create(SrcBitmap.Width, ColorLookup);
        dmSierra:
          Ditherer := TSierraDitherer.Create(SrcBitmap.Width, ColorLookup);
        dmJaJuNI:
          Ditherer := TJaJuNIDitherer.Create(SrcBitmap.Width, ColorLookup);
        dmSteveArche:
          Ditherer := TSteveArcheDitherer.Create(SrcBitmap.Width, ColorLookup);
        dmBurkes:
          Ditherer := TBurkesDitherer.Create(SrcBitmap.Width, ColorLookup);
      else
        exit;
      end;

      DIBResult := TDIBWriter.Create(DstBitmap, xpf8bit, SrcBitmap.Width, SrcBitmap.Height,
        Palette);
      // Process the image
      Row := 0;
      while (Row < SrcBitmap.Height) do
      begin
        SrcScanline := DIBSource.Scanline[Row];
        DstScanline := DIBResult.Scanline[Row];
        Src := pointer(longInt(SrcScanLine) + Ditherer.Column*sizeof(TxRGBTriple));
        Dst := pointer(longInt(DstScanLine) + Ditherer.Column);

        while (Ditherer.Column < Ditherer.Width) and (Ditherer.Column >= 0) do
        begin
          BGR := Src^;
          // Dither and map a single pixel
          Dst^ := Ditherer.Dither(BGR.rgbtRed, BGR.rgbtGreen, BGR.rgbtBlue,
            BGR.rgbtRed, BGR.rgbtGreen, BGR.rgbtBlue);

          inc(Src, Ditherer.Direction);
          inc(Dst, Ditherer.Direction);
        end;

        Inc(Row);
        Ditherer.NextLine;
      end;
    finally
      if (ColorLookup <> nil) then
        ColorLookup.Free;
      if (Ditherer <> nil) then
        Ditherer.Free;

      if (DIBResult <> nil) then
        DIBResult.Free;
      if (DIBSource <> nil) then
        DIBSource.Free;

      // Must delete palette after TDIBWriter since TDIBWriter uses palette
      if (Palette <> 0) then
        xDeleteObject(Palette);
    end;
  {$endif}
end;



//------------------------------------------------------------
// Spliting
//------------------------------------------------------------

// -----------------------------------------------------------------------------
//
//			GetPlane
//
// -----------------------------------------------------------------------------
procedure GetPlane(SrcBitmap,DestBitmap:TXBitmap;Plane:Integer);
 var
   i        :  INTEGER;
   j        :  INTEGER;
   rowRGB   :  pRGBarray;
   rowD     :  pRGBarray;
   HSV      :  THSVTriple;
   HSL      :  THSLTriple;
   BMPIntfc :  TBMPInterface1;
begin

  if not CheckPixelFormat(SrcBitmap,xpf24bit) then Exit;

  BMPIntfc:=TBMPInterface1.Create(SrcBitmap,DestBitmap,xpf24Bit,0,True);
  try
    for j:=0 to BMPIntfc.Height-1 do
    begin

      RowRGB:=BmpIntfc.SourceScanLine[j];
      RowD:=BmpIntfc.DestScanLine[j];

      for i := BMPIntfc.Width-1 downto 0 DO
      begin
        case Plane of
        1:begin
            rowD[i].rgbtRed   := rowRGB[i].rgbtRed;
            rowD[i].rgbtGreen := 0;
            rowD[i].rgbtBlue  := 0
          end;
        2:begin
            rowD[i].rgbtRed   := 0;
            rowD[i].rgbtGreen := rowRGB[i].rgbtGreen;
            rowD[i].rgbtBlue  := 0
          end;
        3:begin
            rowD[i].rgbtRed   := 0;
            rowD[i].rgbtGreen := 0;
            rowD[i].rgbtBlue  := rowRGB[i].rgbtBlue;
          end;
        4:begin
            RGBtoHSV(rowRGB[i],  HSV);
            HSV.hsvHue := (HSV.hsvHue * 255) DIV 359;
            rowD[i].rgbtRed   := HSV.hsvHue;
            rowD[i].rgbtGreen := HSV.hsvHue;
            rowD[i].rgbtBlue  := HSV.hsvHue;
          end;
        5:begin
            RGBtoHSV(rowRGB[i], HSV);
            rowD[i].rgbtRed   := HSV.hsvSaturation;
            rowD[i].rgbtGreen := HSV.hsvSaturation;
            rowD[i].rgbtBlue  := HSV.hsvSaturation;
          end;
        6:begin
            RGBtoHSV(rowRGB[i], HSV);
            rowD[i].rgbtRed   := HSV.hsvValue;
            rowD[i].rgbtGreen := HSV.hsvValue;
            rowD[i].rgbtBlue  := HSV.hsvValue;
          end;
        7:begin
            RGBtoHSL(rowRGB[i], HSL);
            rowD[i].rgbtRed   := HSL.hslLightness ;
            rowD[i].rgbtGreen := HSL.hslLightness;
            rowD[i].rgbtBlue  := HSL.hslLightness;
          end;
        end;
      end;

    end;
    finally
      BMPIntfc.Free;
    end;
end;

// -----------------------------------------------------------------------------
//
//			GetRedPlane
//
// -----------------------------------------------------------------------------
procedure GetRedPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,1);
end;

// -----------------------------------------------------------------------------
//
//			GetGreenPlane
//
// -----------------------------------------------------------------------------
procedure GetGreenPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,2);
end;

// -----------------------------------------------------------------------------
//
//			GetBluePlane
//
// -----------------------------------------------------------------------------
procedure GetBluePlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,3);
end;

// -----------------------------------------------------------------------------
//
//			GetHuePlane
//
// -----------------------------------------------------------------------------
procedure GetHuePlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,4);
end;

// -----------------------------------------------------------------------------
//
//			GetSatPlane
//
// -----------------------------------------------------------------------------
procedure GetSatPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,5);
end;

// -----------------------------------------------------------------------------
//
//			GetValPlane
//
// -----------------------------------------------------------------------------
procedure GetValPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,6);
end;

// -----------------------------------------------------------------------------
//
//			GetLigPlane
//
// -----------------------------------------------------------------------------
procedure GetLigPlane(SrcBitmap:TXBitmap;DestBitmap:TXBitmap);
begin
  GetPlane(SrcBitmap,DestBitmap,7);
end;


end.
